use std::collections::HashMap;

use bumpalo::collections::Vec as BumpVec;
use bumpalo::collections::{CollectIn, String as BumpString};
use bumpalo::Bump;
use chrono::NaiveDateTime;
use chrono::{Datelike, Timelike};
#[cfg(not(feature = "regex-lite"))]
use regex::Regex;
#[cfg(feature = "regex-lite")]
use regex_lite::Regex;
use rust_decimal::prelude::{FromPrimitive, ToPrimitive};
use rust_decimal::{Decimal, MathematicalOps};
use rust_decimal_macros::dec;

use crate::compiler::{Opcode, TypeCheckKind, TypeConversionKind};
use crate::variable::ToVariable;
use crate::variable::Variable;
use crate::variable::Variable::*;
use crate::vm::error::VMError::*;
use crate::vm::error::VMResult;
use crate::vm::helpers::{date_time, date_time_end_of, date_time_start_of, time};
use crate::vm::variable::IntervalObject;

pub(crate) const NULL_VAR: &'static Variable = &Null;

#[derive(Debug)]
pub struct Scope<'arena> {
    array: &'arena Variable<'arena>,
    len: usize,
    iter: usize,
    count: usize,
}

#[derive(Debug)]
pub struct VM<'arena> {
    scopes: Vec<Scope<'arena>>,
    stack: Vec<&'arena Variable<'arena>>,
}

impl<'arena> VM<'arena> {
    pub fn new() -> Self {
        Self {
            scopes: Default::default(),
            stack: Default::default(),
        }
    }

    pub fn run(
        &mut self,
        bytecode: &[Opcode<'arena>],
        bump: &'arena Bump,
        env: &Variable<'arena>,
    ) -> VMResult<&Variable> {
        self.stack.clear();
        self.scopes.clear();

        let s = VMInner::new(bytecode, &mut self.stack, &mut self.scopes, bump).run(env);
        Ok(s?)
    }
}

struct VMInner<'arena, 'parent_ref, 'bytecode_ref> {
    scopes: &'parent_ref mut Vec<Scope<'arena>>,
    stack: &'parent_ref mut Vec<&'arena Variable<'arena>>,
    bytecode: &'bytecode_ref [Opcode<'arena>],
    bump: &'arena Bump,
    ip: usize,
}

impl<'arena, 'parent_ref, 'bytecode_ref> VMInner<'arena, 'parent_ref, 'bytecode_ref> {
    pub fn new(
        bytecode: &'bytecode_ref [Opcode<'arena>],
        stack: &'parent_ref mut Vec<&'arena Variable<'arena>>,
        scopes: &'parent_ref mut Vec<Scope<'arena>>,
        bump: &'arena Bump,
    ) -> Self {
        Self {
            scopes,
            stack,
            ip: 0,
            bytecode,
            bump,
        }
    }

    fn push(&mut self, var: Variable<'arena>) {
        self.stack.push(self.bump.alloc(var));
    }

    fn pop(&mut self) -> VMResult<&'arena Variable<'arena>> {
        self.stack.pop().ok_or_else(|| StackOutOfBounds {
            stack: format!("{:?}", self.stack),
        })
    }

    fn push_ref(&mut self, var: &'arena Variable<'arena>) {
        self.stack.push(var);
    }

    pub fn run(&mut self, env: &Variable<'arena>) -> VMResult<&'arena Variable<'arena>> {
        if self.ip != 0 {
            self.ip = 0;
        }

        while self.ip < self.bytecode.len() {
            let op = self
                .bytecode
                .get(self.ip)
                .ok_or_else(|| OpcodeOutOfBounds {
                    bytecode: format!("{:?}", self.bytecode),
                    index: self.ip,
                })?;

            self.ip += 1;

            match op {
                Opcode::Push(v) => {
                    self.push(v.clone_in(self.bump));
                }
                Opcode::Pop => {
                    self.pop()?;
                }
                Opcode::Rot => {
                    let b = self.stack.len() - 1;
                    let a = self.stack.len() - 2;
                    self.stack.swap(a, b);
                }
                Opcode::Fetch => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (Object(o), String(s)) => {
                            self.push_ref(o.get(*s).unwrap_or(&NULL_VAR));
                        }
                        (Array(arr), Number(n)) => self.push_ref(
                            arr.get(n.to_usize().ok_or_else(|| OpcodeErr {
                                opcode: "Fetch".into(),
                                message: "Failed to convert to usize".into(),
                            })?)
                            .unwrap_or(&NULL_VAR),
                        ),
                        (String(str), Number(n)) => {
                            let index = n.to_usize().ok_or_else(|| OpcodeErr {
                                opcode: "Fetch".into(),
                                message: "Failed to convert to usize".into(),
                            })?;

                            if let Some(slice) = str.get(index..index + 1) {
                                self.push(String(self.bump.alloc_str(slice)));
                            } else {
                                self.push_ref(&NULL_VAR)
                            };
                        }
                        _ => self.push_ref(&NULL_VAR),
                    }
                }
                Opcode::FetchEnv(f) => match env {
                    Object(o) => match o.get(*f) {
                        None => self.push_ref(&NULL_VAR),
                        Some(v) => self.push(v.clone_in(self.bump)),
                    },
                    Null => self.push_ref(NULL_VAR),
                    _ => {
                        return Err(OpcodeErr {
                            opcode: "FetchEnv".into(),
                            message: "Unsupported type".into(),
                        });
                    }
                },
                Opcode::FetchRootEnv => {
                    self.push(env.clone_in(self.bump));
                }
                Opcode::Negate => {
                    let a = self.pop()?;
                    match a {
                        Number(n) => {
                            self.push(Number(-*n));
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Negate".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Not => {
                    let a = self.pop()?;
                    match a {
                        Bool(b) => self.push(Bool(!(*b))),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Not".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Equal => {
                    let b = self.pop()?;
                    let a = self.pop()?;
                    match (a, b) {
                        (Number(a), Number(b)) => {
                            self.push(Bool(a == b));
                        }
                        (Bool(a), Bool(b)) => {
                            self.push(Bool(a == b));
                        }
                        (String(a), String(b)) => {
                            self.push(Bool(a == b));
                        }
                        (Null, Null) => {
                            self.push(Bool(true));
                        }
                        _ => {
                            self.push(Bool(false));
                        }
                    }
                }
                Opcode::Jump(j) => self.ip += j,
                Opcode::JumpIfTrue(j) => {
                    let a = self.stack.last().ok_or_else(|| OpcodeErr {
                        opcode: "JumpIfTrue".into(),
                        message: "Undefined object key".into(),
                    })?;
                    match a {
                        Bool(a) => {
                            if *a {
                                self.ip += j;
                            }
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "JumpIfTrue".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::JumpIfFalse(j) => {
                    let a = self.stack.last().ok_or_else(|| OpcodeErr {
                        opcode: "JumpIfFalse".into(),
                        message: "Empty array".into(),
                    })?;

                    match a {
                        Bool(a) => {
                            if !*a {
                                self.ip += j;
                            }
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "JumpIfFalse".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::JumpBackward(j) => {
                    self.ip -= j;
                }
                Opcode::In => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (Number(a), Array(arr)) => {
                            let is_in = arr.iter().any(|b| match b {
                                Number(b) => a == b,
                                _ => false,
                            });

                            self.push(Bool(is_in));
                        }
                        (Number(v), Object(_)) => {
                            let interval =
                                IntervalObject::try_from_object(b).ok_or_else(|| OpcodeErr {
                                    opcode: "In".into(),
                                    message: "Failed to deconstruct interval".into(),
                                })?;

                            match (interval.left, interval.right) {
                                (Number(l), Number(r)) => {
                                    let mut is_open = false;

                                    let first = match interval.left_bracket {
                                        "[" => l <= v,
                                        "(" => l < v,
                                        "]" => {
                                            is_open = true;
                                            l >= v
                                        }
                                        ")" => {
                                            is_open = true;
                                            l > v
                                        }
                                        _ => {
                                            return Err(OpcodeErr {
                                                opcode: "In".into(),
                                                message: "Unsupported bracket".into(),
                                            });
                                        }
                                    };

                                    let second = match interval.right_bracket {
                                        "]" => r >= v,
                                        ")" => r > v,
                                        "[" => r <= v,
                                        "(" => r < v,
                                        _ => {
                                            return Err(OpcodeErr {
                                                opcode: "In".into(),
                                                message: "Unsupported bracket".into(),
                                            });
                                        }
                                    };

                                    let open_stmt = is_open && (first || second);
                                    let closed_stmt = !is_open && first && second;

                                    self.push(Bool(open_stmt || closed_stmt));
                                }
                                _ => {
                                    return Err(OpcodeErr {
                                        opcode: "In".into(),
                                        message: "Unsupported type".into(),
                                    });
                                }
                            }
                        }
                        (String(a), Array(arr)) => {
                            let is_in = arr.iter().any(|b| match b {
                                String(b) => a == b,
                                _ => false,
                            });

                            self.push(Bool(is_in));
                        }
                        (Bool(a), Array(arr)) => {
                            let is_in = arr.iter().any(|b| match b {
                                Bool(b) => a == b,
                                _ => false,
                            });

                            self.push(Bool(is_in));
                        }
                        (Null, Array(arr)) => {
                            let is_in = arr.iter().any(|b| match b {
                                Null => true,
                                _ => false,
                            });

                            self.push(Bool(is_in));
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "In".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Less => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (Number(a), Number(b)) => self.push(Bool(*a < *b)),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Less".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::More => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (Number(a), Number(b)) => self.push(Bool(*a > *b)),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "More".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::LessOrEqual => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (Number(a), Number(b)) => self.push(Bool(*a <= *b)),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "LessOrEqual".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::MoreOrEqual => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (Number(a), Number(b)) => self.push(Bool(*a >= *b)),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "MoreOrEqual".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Abs => {
                    let a = self.pop()?;

                    match a {
                        Number(a) => self.push(Number(a.abs())),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Abs".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Round => {
                    let var = self.pop()?;

                    match var {
                        Number(a) => self.push(Number(a.round())),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Round".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Ceil => {
                    let var = self.pop()?;

                    match var {
                        Number(a) => self.push(Number(a.ceil())),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Ceil".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Floor => {
                    let var = self.pop()?;

                    match var {
                        Number(a) => self.push(Number(a.floor())),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Floor".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Random => {
                    let var = self.pop()?;
                    match var {
                        Number(a) => {
                            let upper_range = a.round().to_i64().ok_or_else(|| OpcodeErr {
                                opcode: "Random".into(),
                                message: "Failed to determine upper range".into(),
                            })?;

                            let random_number = fastrand::i64(0..=upper_range);
                            self.push(Number(Decimal::from(random_number)));
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Random".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Average => {
                    let var = self.pop()?;

                    match var {
                        Array(arr) => {
                            let mut sum = Decimal::ZERO;
                            arr.iter().try_for_each(|a| match a {
                                Number(a) => {
                                    sum += a;
                                    Ok(())
                                }
                                _ => Err(OpcodeErr {
                                    opcode: "Average".into(),
                                    message: "Invalid array value".into(),
                                }),
                            })?;

                            let avg = sum / Decimal::from(arr.len());
                            self.push(Number(avg));
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Average".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Median => {
                    let Array(arr) = self.pop()? else {
                        return Err(OpcodeErr {
                            opcode: "Median".into(),
                            message: "Unsupported type".into(),
                        });
                    };

                    let mut num_arr = arr
                        .iter()
                        .map(|n| match n {
                            Number(num) => Ok(num),
                            _ => Err(OpcodeErr {
                                opcode: "Median".into(),
                                message: "Unsupported type".into(),
                            }),
                        })
                        .collect::<Result<Vec<_>, _>>()?;

                    if num_arr.len() == 0 {
                        return Err(OpcodeErr {
                            opcode: "Median".into(),
                            message: "Array is empty".into(),
                        });
                    }

                    num_arr.sort();

                    let center = num_arr.len() / 2;
                    if num_arr.len() % 2 == 1 {
                        let center_num = num_arr.get(center).ok_or_else(|| OpcodeErr {
                            opcode: "Median".into(),
                            message: "Array out of bounds".into(),
                        })?;

                        self.push(Number(**center_num));
                    } else {
                        let center_left = num_arr.get(center - 1).ok_or_else(|| OpcodeErr {
                            opcode: "Median".into(),
                            message: "Array out of bounds".into(),
                        })?;

                        let center_right = num_arr.get(center).ok_or_else(|| OpcodeErr {
                            opcode: "Median".into(),
                            message: "Array out of bounds".into(),
                        })?;

                        let median = ((**center_left) + (**center_right)) / dec!(2);
                        self.push(Number(median));
                    }
                }
                Opcode::Mode => {
                    let Array(arr) = self.pop()? else {
                        return Err(OpcodeErr {
                            opcode: "Mode".into(),
                            message: "Unsupported type".into(),
                        });
                    };

                    let num_arr = arr
                        .iter()
                        .map(|n| match n {
                            Number(num) => Ok(num),
                            _ => Err(OpcodeErr {
                                opcode: "Mode".into(),
                                message: "Unsupported type".into(),
                            }),
                        })
                        .collect::<Result<Vec<_>, _>>()?;

                    if num_arr.len() == 0 {
                        return Err(OpcodeErr {
                            opcode: "Mode".into(),
                            message: "Array is empty".into(),
                        });
                    }

                    let mut map = HashMap::new();
                    num_arr.iter().for_each(|n| {
                        let count = map.entry(**n).or_insert(0);
                        *count += 1;
                    });

                    let maybe_mode = map
                        .iter()
                        .max_by_key(|&(_, count)| *count)
                        .map(|(val, _)| val);

                    let mode = maybe_mode.ok_or_else(|| OpcodeErr {
                        opcode: "Mode".into(),
                        message: "Failed to find most common element".into(),
                    })?;

                    self.push(Number(*mode));
                }
                Opcode::Min => {
                    let var = self.pop()?;

                    match var {
                        Array(arr) => {
                            let first_item = arr.get(0).ok_or_else(|| OpcodeErr {
                                opcode: "Min".into(),
                                message: "Empty array".into(),
                            })?;

                            let mut min: Decimal = match first_item {
                                Number(a) => *a,
                                _ => {
                                    return Err(OpcodeErr {
                                        opcode: "Min".into(),
                                        message: "Unsupported array value".into(),
                                    });
                                }
                            };

                            arr.iter().try_for_each(|a| match a {
                                Number(a) => {
                                    if *a < min {
                                        min = *a;
                                    }
                                    Ok(())
                                }
                                _ => Err(OpcodeErr {
                                    opcode: "Min".into(),
                                    message: "Unsupported array value".into(),
                                }),
                            })?;

                            self.push(Number(min));
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Min".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Max => {
                    let var = self.pop()?;

                    match var {
                        Array(arr) => {
                            let first_item = arr.get(0).ok_or_else(|| OpcodeErr {
                                opcode: "Max".into(),
                                message: "Empty array".into(),
                            })?;

                            let mut max: Decimal = match first_item {
                                Number(a) => *a,
                                _ => {
                                    return Err(OpcodeErr {
                                        opcode: "Max".into(),
                                        message: "Unsupported array value".into(),
                                    });
                                }
                            };

                            arr.iter().try_for_each(|a| match a {
                                Number(a) => {
                                    if *a > max {
                                        max = *a;
                                    }
                                    Ok(())
                                }
                                _ => Err(OpcodeErr {
                                    opcode: "Max".into(),
                                    message: "Unsupported array value".into(),
                                }),
                            })?;

                            self.push(Number(max));
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Max".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Sum => {
                    let var = self.pop()?;

                    match var {
                        Array(arr) => {
                            let mut sum = Decimal::ZERO;
                            arr.iter().try_for_each(|a| match a {
                                Number(a) => {
                                    sum += a;
                                    Ok(())
                                }
                                _ => Err(OpcodeErr {
                                    opcode: "Sum".into(),
                                    message: "Unsupported array value".into(),
                                }),
                            })?;

                            self.push(Number(sum));
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Sum".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Add => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (Number(a), Number(b)) => self.push(Number(a + b)),
                        (String(a), String(b)) => {
                            let mut str1 =
                                BumpString::with_capacity_in(a.len() + b.len(), self.bump);
                            str1.push_str(a);
                            str1.push_str(b);

                            self.push(String(str1.into_bump_str()));
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Add".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Subtract => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (Number(a), Number(b)) => self.push(Number(a - b)),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Subtract".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Multiply => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (Number(a), Number(b)) => self.push(Number(a * b)),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Multiply".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Divide => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (Number(a), Number(b)) => self.push(Number(a / b)),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Divide".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Modulo => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (Number(a), Number(b)) => self.push(Number(a % b)),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Modulo".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Exponent => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (Number(a), Number(b)) => {
                            self.push(Number(a.powd(*b)));
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Exponent".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Interval {
                    left_bracket,
                    right_bracket,
                } => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (Number(_), Number(_)) => {
                            let interval = IntervalObject {
                                left_bracket,
                                right_bracket,
                                left: a,
                                right: b,
                            };

                            self.push(interval.to_variable(self.bump).unwrap());
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Interval".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Uppercase => {
                    let a = self.pop()?;

                    match a {
                        String(a) => {
                            let str = a.to_uppercase();
                            self.push(String(self.bump.alloc_str(&str)));
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Uppercase".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Lowercase => {
                    let a = self.pop()?;

                    match a {
                        String(a) => {
                            let str = a.to_lowercase();
                            self.push(String(self.bump.alloc_str(&str)));
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Lowercase".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Contains => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (String(a), String(b)) => {
                            self.push(Bool(a.contains(b)));
                        }
                        _ => match a {
                            Array(arr) => {
                                let is_in = arr.iter().any(|a| match (a, b) {
                                    (Number(a), Number(b)) => a == b,
                                    (String(a), String(b)) => a == b,
                                    (Bool(a), Bool(b)) => a == b,
                                    (Null, Null) => true,
                                    _ => false,
                                });

                                self.push(Bool(is_in));
                            }
                            _ => {
                                return Err(OpcodeErr {
                                    opcode: "Contains".into(),
                                    message: "Unsupported type".into(),
                                });
                            }
                        },
                    }
                }
                Opcode::Keys => {
                    let current = self.pop()?;

                    match current {
                        Array(arr) => {
                            let indices: BumpVec<Variable> = arr
                                .iter()
                                .enumerate()
                                .map(|(index, _)| Number(index.into()))
                                .collect_in(self.bump);

                            self.push(Array(indices));
                        }
                        Object(obj) => {
                            let keys: BumpVec<Variable> = obj
                                .iter()
                                .map(|(key, _)| String(self.bump.alloc_str(key)))
                                .collect_in(self.bump);

                            self.push(Array(keys));
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Keys".into(),
                                message: "Unsupported type".into(),
                            })
                        }
                    }
                }
                Opcode::StartsWith => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (String(a), String(b)) => {
                            self.push(Bool(a.starts_with(b)));
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "StartsWith".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::EndsWith => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    match (a, b) {
                        (String(a), String(b)) => {
                            self.push(Bool(a.ends_with(b)));
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "EndsWith".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Matches => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    let (String(a), String(b)) = (a, b) else {
                        return Err(OpcodeErr {
                            opcode: "Matches".into(),
                            message: "Unsupported type".into(),
                        });
                    };

                    let regex = Regex::new(b).map_err(|_| OpcodeErr {
                        opcode: "Matches".into(),
                        message: "Invalid regular expression".into(),
                    })?;

                    self.push(Bool(regex.is_match(a)));
                }
                Opcode::FuzzyMatch => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    let String(b) = b else {
                        return Err(OpcodeErr {
                            opcode: "FuzzyMatch".into(),
                            message: "Unsupported type".into(),
                        });
                    };

                    match a {
                        String(a) => {
                            let sim = strsim::normalized_damerau_levenshtein(a, b);
                            // This is okay, as NDL will return [0, 1]
                            self.push(Number(Decimal::from_f64(sim).unwrap_or(dec!(0))));
                        }
                        Array(a) => {
                            let mut sims = BumpVec::with_capacity_in(a.len(), &self.bump);
                            for v in a.iter() {
                                let String(s) = v else {
                                    return Err(OpcodeErr {
                                        opcode: "FuzzyMatch".into(),
                                        message: "Unsupported type".into(),
                                    });
                                };

                                let sim =
                                    Decimal::from_f64(strsim::normalized_damerau_levenshtein(s, b))
                                        .unwrap_or(dec!(0));
                                sims.push(Number(sim));
                            }

                            self.push(Array(sims))
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "FuzzyMatch".into(),
                                message: "Unsupported type".into(),
                            })
                        }
                    }
                }
                Opcode::Extract => {
                    let b = self.pop()?;
                    let a = self.pop()?;

                    let (String(a), String(b)) = (a, b) else {
                        return Err(OpcodeErr {
                            opcode: "Matches".into(),
                            message: "Unsupported type".into(),
                        });
                    };

                    let regex = Regex::new(b).map_err(|_| OpcodeErr {
                        opcode: "Matches".into(),
                        message: "Invalid regular expression".into(),
                    })?;

                    let captures = regex
                        .captures(a)
                        .map(|capture| {
                            capture
                                .iter()
                                .map(|c| c.map(|c| c.as_str()))
                                .filter_map(|c| c)
                                .map(|s| String(self.bump.alloc_str(s)))
                                .collect_in::<BumpVec<_>>(self.bump)
                        })
                        .unwrap_or_else(|| BumpVec::new_in(self.bump));

                    self.push(Array(captures));
                }
                Opcode::DateManipulation(operation) => {
                    let timestamp = self.pop()?;

                    let time: NaiveDateTime = timestamp.try_into()?;
                    let var = match *operation {
                        "year" => Number(time.year().into()),
                        "dayOfWeek" => Number(time.weekday().number_from_monday().into()),
                        "dayOfMonth" => Number(time.day().into()),
                        "dayOfYear" => Number(time.ordinal().into()),
                        "weekOfYear" => Number(time.iso_week().week().into()),
                        "monthOfYear" => Number(time.month().into()),
                        "monthString" => {
                            String(self.bump.alloc_str(&time.format("%b").to_string()))
                        }
                        "weekdayString" => String(self.bump.alloc_str(&time.weekday().to_string())),
                        "dateString" => String(self.bump.alloc_str(&time.to_string())),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "DateManipulation".into(),
                                message: "Unsupported operation".into(),
                            });
                        }
                    };

                    self.push(var);
                }
                Opcode::DateFunction(name) => {
                    let unit_var = self.pop()?;
                    let timestamp = self.pop()?;

                    let date_time: NaiveDateTime = timestamp.try_into()?;
                    let String(unit_name) = *unit_var else {
                        return Err(OpcodeErr {
                            opcode: "DateFunction".into(),
                            message: "Unknown date function".into(),
                        });
                    };

                    let s = match *name {
                        "startOf" => date_time_start_of(date_time, unit_name.try_into()?),
                        "endOf" => date_time_end_of(date_time, unit_name.try_into()?),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "DateManipulation".into(),
                                message: "Unsupported operation".into(),
                            });
                        }
                    }
                    .ok_or_else(|| OpcodeErr {
                        opcode: "DateFunction".into(),
                        message: "Failed to run DateFunction".into(),
                    })?;

                    self.push(Number(s.timestamp().into()));
                }
                Opcode::Slice => {
                    let from_var = self.pop()?;
                    let to_var = self.pop()?;
                    let current = self.pop()?;

                    match (from_var, to_var) {
                        (Number(f), Number(t)) => {
                            let from = f.to_usize().ok_or_else(|| OpcodeErr {
                                opcode: "Slice".into(),
                                message: "Failed to get range from".into(),
                            })?;
                            let to = t.to_usize().ok_or_else(|| OpcodeErr {
                                opcode: "Slice".into(),
                                message: "Failed to get range to".into(),
                            })?;

                            match current {
                                Array(arr) => {
                                    let slice = arr.get(from..=to).ok_or_else(|| OpcodeErr {
                                        opcode: "Slice".into(),
                                        message: "Index out of range".into(),
                                    })?;

                                    self.push(Array(BumpVec::from_iter_in(
                                        slice.iter().map(|v| v.clone_in(self.bump)),
                                        self.bump,
                                    )));
                                }
                                String(s) => {
                                    let slice = s.get(from..=to).ok_or_else(|| OpcodeErr {
                                        opcode: "Slice".into(),
                                        message: "Index out of range".into(),
                                    })?;

                                    self.push(String(slice));
                                }
                                _ => {
                                    return Err(OpcodeErr {
                                        opcode: "Slice".into(),
                                        message: "Unsupported type".into(),
                                    });
                                }
                            }
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Slice".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::Array => {
                    let size = self.pop()?;

                    let Number(s) = size else {
                        return Err(OpcodeErr {
                            opcode: "Array".into(),
                            message: "Unsupported type".into(),
                        });
                    };

                    let to = s.round().to_usize().ok_or_else(|| OpcodeErr {
                        opcode: "Array".into(),
                        message: "Failed to extract argument".into(),
                    })?;

                    let mut arr = BumpVec::with_capacity_in(to, &self.bump);
                    for _ in 0..to {
                        arr.push(self.pop()?.clone_in(self.bump));
                    }
                    arr.reverse();

                    self.push(Array(arr));
                }
                Opcode::Len => {
                    let current = self.stack.last().ok_or_else(|| OpcodeErr {
                        opcode: "Len".into(),
                        message: "Empty stack".into(),
                    })?;

                    let len = match current {
                        String(s) => s.len(),
                        Array(s) => s.len(),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Len".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    };

                    self.push(Number(len.into()));
                }
                Opcode::Flatten => {
                    let current = self.pop()?;
                    let Array(arr) = current else {
                        return Err(OpcodeErr {
                            opcode: "Flatten".into(),
                            message: "Unsupported type".into(),
                        });
                    };

                    let mut flat_arr = BumpVec::new_in(&self.bump);
                    flat_arr.reserve(arr.len());

                    arr.iter().for_each(|v| match v {
                        Array(arr) => arr
                            .iter()
                            .for_each(|v| flat_arr.push(v.clone_in(self.bump))),
                        _ => flat_arr.push(v.clone_in(self.bump)),
                    });

                    self.push(Array(flat_arr));
                }
                Opcode::ParseDateTime => {
                    let a = self.pop()?;
                    let ts = match a {
                        String(a) => date_time(a)?.timestamp(),
                        Number(a) => a.to_i64().ok_or_else(|| OpcodeErr {
                            opcode: "ParseDateTime".into(),
                            message: "Number overflow".into(),
                        })?,
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "ParseDateTime".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    };

                    self.push(Number(ts.into()));
                }
                Opcode::ParseTime => {
                    let a = self.pop()?;
                    let ts = match a {
                        String(a) => time(a)?.num_seconds_from_midnight(),
                        Number(a) => a.to_u32().ok_or_else(|| OpcodeErr {
                            opcode: "ParseTime".into(),
                            message: "Number overflow".into(),
                        })?,
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "ParseTime".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    };

                    self.push(Number(ts.into()));
                }
                Opcode::ParseDuration => {
                    let a = self.pop()?;

                    let dur = match a {
                        String(a) => humantime::parse_duration(a)
                            .map_err(|_| ParseDateTimeErr {
                                timestamp: a.to_string(),
                            })?
                            .as_secs(),
                        Number(n) => n.to_u64().ok_or_else(|| OpcodeErr {
                            opcode: "ParseDuration".into(),
                            message: "Number overflow".into(),
                        })?,
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "ParseDuration".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    };

                    self.push(Number(dur.into()));
                }
                Opcode::TypeCheck(check) => {
                    let var = self.pop()?;

                    let is_equal = match (check, var) {
                        (TypeCheckKind::Numeric, String(str)) => {
                            Decimal::from_str_exact(str).is_ok()
                        }
                        (TypeCheckKind::Numeric, Number(_)) => true,
                        (TypeCheckKind::Numeric, _) => false,
                    };

                    self.push(Bool(is_equal));
                }
                Opcode::TypeConversion(conversion) => {
                    let var = self.pop()?;

                    let make_string = |val: &str| self.bump.alloc(String(self.bump.alloc_str(val)));

                    let converted_var = match (conversion, var) {
                        (TypeConversionKind::String, String(_)) => var,
                        (TypeConversionKind::String, Number(num)) => {
                            make_string(num.to_string().as_str())
                        }
                        (TypeConversionKind::String, Bool(v)) => {
                            make_string(v.to_string().as_str())
                        }
                        (TypeConversionKind::String, Null) => make_string("null"),
                        (TypeConversionKind::String, _) => {
                            return Err(OpcodeErr {
                                opcode: "TypeConversion".into(),
                                message: format!(
                                    "Type {} cannot be converted to string",
                                    var.type_name()
                                ),
                            });
                        }
                        (TypeConversionKind::Number, String(str)) => {
                            let parsed_number =
                                Decimal::from_str_exact(str).map_err(|_| OpcodeErr {
                                    opcode: "TypeConversion".into(),
                                    message: "Failed to parse string to number".into(),
                                })?;

                            self.bump.alloc(Number(parsed_number))
                        }
                        (TypeConversionKind::Number, Number(_)) => var,
                        (TypeConversionKind::Number, Bool(v)) => {
                            let number = if *v { dec!(1) } else { dec!(0) };
                            self.bump.alloc(Number(number))
                        }
                        (TypeConversionKind::Number, _) => {
                            return Err(OpcodeErr {
                                opcode: "TypeConversion".into(),
                                message: format!(
                                    "Type {} cannot be converted to number",
                                    var.type_name()
                                ),
                            });
                        }
                        (TypeConversionKind::Bool, Number(n)) => {
                            self.bump.alloc(Bool(!n.is_zero()))
                        }
                        (TypeConversionKind::Bool, String(s)) => {
                            let value = match *s {
                                "true" => true,
                                "false" => false,
                                _ => s.is_empty(),
                            };

                            self.bump.alloc(Bool(value))
                        }
                        (TypeConversionKind::Bool, Bool(_)) => var,
                        (TypeConversionKind::Bool, Null) => self.bump.alloc(Bool(false)),
                        (TypeConversionKind::Bool, Object(_) | Array(_)) => {
                            self.bump.alloc(Bool(true))
                        }
                    };

                    self.push_ref(converted_var);
                }
                Opcode::JumpIfEnd(j) => {
                    let scope = self.scopes.last().ok_or_else(|| OpcodeErr {
                        opcode: "JumpIfEnd".into(),
                        message: "Empty stack".into(),
                    })?;

                    if scope.iter >= scope.len {
                        self.ip += j;
                    }
                }
                Opcode::IncrementIt => {
                    let scope = self.scopes.last_mut().ok_or_else(|| OpcodeErr {
                        opcode: "IncrementIt".into(),
                        message: "Empty scope".into(),
                    })?;

                    scope.iter += 1;
                }
                Opcode::IncrementCount => {
                    let scope = self.scopes.last_mut().ok_or_else(|| OpcodeErr {
                        opcode: "IncrementCount".into(),
                        message: "Empty scope".into(),
                    })?;

                    scope.count += 1;
                }
                Opcode::GetCount => {
                    let scope = self.scopes.last().ok_or_else(|| OpcodeErr {
                        opcode: "GetCount".into(),
                        message: "Empty scope".into(),
                    })?;

                    self.push(Number(scope.count.into()));
                }
                Opcode::GetLen => {
                    let scope = self.scopes.last().ok_or_else(|| OpcodeErr {
                        opcode: "GetLen".into(),
                        message: "Empty scope".into(),
                    })?;

                    self.push(Number(scope.len.into()));
                }
                Opcode::Pointer => {
                    let scope = self.scopes.last().ok_or_else(|| OpcodeErr {
                        opcode: "Pointer".into(),
                        message: "Empty scope".into(),
                    })?;

                    match scope.array {
                        Array(arr) => {
                            let variable = arr.get(scope.iter).ok_or_else(|| OpcodeErr {
                                opcode: "Pointer".into(),
                                message: "Scope array out of bounds".into(),
                            })?;

                            self.push_ref(variable);
                        }
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Pointer".into(),
                                message: "Unsupported scope type".into(),
                            });
                        }
                    }
                }
                Opcode::Begin => {
                    let a = self.pop()?;
                    match a {
                        Array(arr) => self.scopes.push(Scope {
                            array: a,
                            count: 0,
                            len: arr.len(),
                            iter: 0,
                        }),
                        _ => {
                            return Err(OpcodeErr {
                                opcode: "Begin".into(),
                                message: "Unsupported type".into(),
                            });
                        }
                    }
                }
                Opcode::End => {
                    self.scopes.pop();
                }
            }
        }

        self.pop()
    }
}
